##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""ReportLab fixups.
"""
__docformat__ = "reStructuredText"
import copy
from reportlab.pdfbase import pdfform, pdfmetrics, ttfonts
from reportlab.pdfbase.pdfpattern import PDFPattern
from reportlab.graphics import testshapes
from reportlab.lib import fonts

_ps2tt_map_original = copy.deepcopy(fonts._ps2tt_map)
_tt2ps_map_original = copy.deepcopy(fonts._tt2ps_map)

def resetPdfForm():
    pdfform.PDFDOCENC = PDFPattern(pdfform.PDFDocEncodingPattern)
    pdfform.ENCODING = PDFPattern(
        pdfform.EncodingPattern, PDFDocEncoding=pdfform.PDFDOCENC)
    pdfform.GLOBALFONTSDICTIONARY = pdfform.FormFontsDictionary()
    pdfform.GLOBALRESOURCES = pdfform.FormResources()
    pdfform.ZADB = PDFPattern(pdfform.ZaDbPattern)

def resetFonts():
    # testshapes._setup registers the Vera fonts every time which is a little
    # slow on all platforms. On Windows it lists the entire system font
    # directory and registers them all which is very slow.
    pdfmetrics.registerFont(ttfonts.TTFont("Vera", "Vera.ttf"))
    pdfmetrics.registerFont(ttfonts.TTFont("VeraBd", "VeraBd.ttf"))
    pdfmetrics.registerFont(ttfonts.TTFont("VeraIt", "VeraIt.ttf"))
    pdfmetrics.registerFont(ttfonts.TTFont("VeraBI", "VeraBI.ttf"))
    for f in ('Times-Roman','Courier','Helvetica','Vera', 'VeraBd', 'VeraIt',
              'VeraBI'):
        if f not in testshapes._FONTS:
            testshapes._FONTS.append(f)
    fonts._ps2tt_map = copy.deepcopy(_ps2tt_map_original)
    fonts._tt2ps_map = copy.deepcopy(_tt2ps_map_original)

def setSideLabels():
    from reportlab.graphics.charts import piecharts
    piecharts.Pie3d.sideLabels = 0
setSideLabels()

from reportlab.rl_config import register_reset
register_reset(resetPdfForm)
register_reset(resetFonts)
del register_reset

# Support more enumeration formats.

from z3c.rml import num2words
from reportlab.lib.sequencer import _type2formatter

_type2formatter.update({
    'l': lambda v: num2words.num2words(v),
    'L': lambda v: num2words.num2words(v).upper(),
    'o': lambda v: num2words.num2words(v, ordinal=True),
    'O': lambda v: num2words.num2words(v, ordinal=True).upper(),
    'r': lambda v: num2words.toOrdinal(v),
    'R': lambda v: num2words.toOrdinal(v).upper(),
})


# Make sure that the counter gets increased for our new formatters as well.

from reportlab.platypus.flowables import ListFlowable, LIIndenter, _LIParams, \
    _computeBulletWidth

def ListFlowable_getContent(self):
    bt = self._bulletType
    value = self._start
    if isinstance(value,(list,tuple)):
        values = value
        value = values[0]
    else:
        values = [value]
    autov = values[0]
    # FIX TO ALLOW ALL FORMATTERS!!!
    inc = int(bt in self._numberStyles)
    if inc:
        try:
            value = int(value)
        except:
            value = 1

    bd = self._bulletDedent
    if bd=='auto':
        align = self._bulletAlign
        dir = self._bulletDir
        if dir=='ltr' and align=='left':
            bd = self._leftIndent
        elif align=='right':
            bd = self._rightIndent
        else:
            #we need to work out the maximum width of any of the labels
            tvalue = value
            maxW = 0
            for d,f in self._flowablesIter():
                if d:
                    maxW = max(maxW,_computeBulletWidth(self,tvalue))
                    if inc: tvalue += inc
                elif isinstance(f,LIIndenter):
                    b = f._bullet
                    if b:
                        if b.bulletType==bt:
                            maxW = max(maxW,_computeBulletWidth(b,b.value))
                            tvalue = int(b.value)
                    else:
                        maxW = max(maxW,_computeBulletWidth(self,tvalue))
                    if inc: tvalue += inc
            if dir=='ltr':
                if align=='right':
                    bd = self._leftIndent - maxW
                else:
                    bd = self._leftIndent - maxW*0.5
            elif align=='left':
                bd = self._rightIndent - maxW
            else:
                bd = self._rightIndent - maxW*0.5

    self._calcBulletDedent = bd

    S = []
    aS = S.append
    i=0
    for d,f in self._flowablesIter():
        if isinstance(f,ListFlowable):
            fstart = f._start
            if isinstance(fstart,(list,tuple)):
                fstart = fstart[0]
            if fstart in values:
                #my kind of ListFlowable
                if f._auto:
                    autov = values.index(autov)+1
                    f._start = values[autov:]+values[:autov]
                    autov = f._start[0]
                    if inc: f._bulletType = autov
                else:
                    autov = fstart
        fparams = {}
        if not i:
            i += 1
            spaceBefore = getattr(self,'spaceBefore',None)
            if spaceBefore is not None:
                fparams['spaceBefore'] = spaceBefore
        if d:
            aS(self._makeLIIndenter(f,bullet=self._makeBullet(value),params=fparams))
            if inc: value += inc
        elif isinstance(f,LIIndenter):
            b = f._bullet
            if b:
                if b.bulletType!=bt:
                    raise ValueError('Included LIIndenter bulletType=%s != OrderedList bulletType=%s' % (b.bulletType,bt))
                value = int(b.value)
            else:
                f._bullet = self._makeBullet(value,params=getattr(f,'params',None))
            if fparams:
                f.__dict__['spaceBefore'] = max(f.__dict__.get('spaceBefore',0),spaceBefore)
            aS(f)
            if inc: value += inc
        elif isinstance(f,_LIParams):
            fparams.update(f.params)
            z = self._makeLIIndenter(f.flowable,bullet=None,params=fparams)
            if f.first:
                if f.value is not None:
                    value = f.value
                    if inc: value = int(value)
                z._bullet = self._makeBullet(value,f.params)
                if inc: value += inc
            aS(z)
        else:
            aS(self._makeLIIndenter(f,bullet=None,params=fparams))

    spaceAfter = getattr(self,'spaceAfter',None)
    if spaceAfter is not None:
        f=S[-1]
        f.__dict__['spaceAfter'] = max(f.__dict__.get('spaceAfter',0),spaceAfter)
    return S

ListFlowable._getContent = ListFlowable_getContent
ListFlowable._numberStyles = _type2formatter.keys()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Page Drawing Related Element Processing
"""
import zope.interface
import reportlab.pdfgen.canvas
from z3c.rml import attr, directive, interfaces, occurence, stylesheet
from z3c.rml import chart, flowable, form, page, special


class IShape(interfaces.IRMLDirectiveSignature):
    """A shape to be drawn on the canvas."""

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'shape.'),
        required=True)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'shape.'),
        required=True)

    fill = attr.Boolean(
        title=u'Fill',
        description=u'A flag to specify whether the shape should be filled.',
        required=False)

    stroke = attr.Boolean(
        title=u'Stroke',
        description=(u"A flag to specify whether the shape's outline should "
                     u"be drawn."),
        required=False)


class CanvasRMLDirective(directive.RMLDirective):
    callable = None
    attrMapping = None

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        getattr(canvas, self.callable)(**kwargs)


class ISaveState(interfaces.IRMLDirectiveSignature):
    """Saves the current canvas state."""

class SaveState(CanvasRMLDirective):
    signature = ISaveState
    callable = 'saveState'


class IRestoreState(interfaces.IRMLDirectiveSignature):
    """Saves the current canvas state."""

class RestoreState(CanvasRMLDirective):
    signature = IRestoreState
    callable = 'restoreState'


class IDrawString(interfaces.IRMLDirectiveSignature):
    """Draws a simple string (left aligned) onto the canvas at the specified
    location."""

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'string.'),
        required=True)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'string.'),
        required=True)

    text = attr.RawXMLContent(
        title=u'Text',
        description=(u'The string/text that is put onto the canvas.'),
        required=True)

class DrawString(CanvasRMLDirective, special.TextFlowables):
    signature = IDrawString
    callable = 'drawString'

    def process(self):
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        kwargs['text'] = self._getText(self.element, canvas).strip()
        getattr(canvas, self.callable)(**kwargs)

class IDrawRightString(IDrawString):
    """Draws a simple string (right aligned) onto the canvas at the specified
    location."""

class DrawRightString(DrawString):
    signature = IDrawRightString
    callable = 'drawRightString'


class IDrawCenteredString(IDrawString):
    """Draws a simple string (centered aligned) onto the canvas at the specified
    location."""

class DrawCenteredString(DrawString):
    signature = IDrawCenteredString
    callable = 'drawCentredString'


class IDrawAlignedString(IDrawString):
    """Draws a simple string (aligned to the pivot character) onto the canvas
    at the specified location."""

    pivotChar = attr.Text(
        title=u'Text',
        description=(u'The string/text that is put onto the canvas.'),
        min_length=1,
        max_length=1,
        default=u'.',
        required=True)

class DrawAlignedString(DrawString):
    signature = IDrawAlignedString
    callable = 'drawAlignedString'


class IEllipse(IShape):
    """Draws an ellipse on the canvas."""

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the ellipse.',
        required=True)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the ellipse.',
        required=True)

class Ellipse(CanvasRMLDirective):
    signature = IEllipse
    callable = 'ellipse'
    attrMapping = {'x': 'x1', 'y': 'y1'}

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        # Convert width and height to end locations
        kwargs['x2'] = kwargs['x1'] + kwargs['width']
        del kwargs['width']
        kwargs['y2'] = kwargs['y1'] + kwargs['height']
        del kwargs['height']
        getattr(canvas, self.callable)(**kwargs)


class ICircle(IShape):
    """Draws a circle on the canvas."""

    radius = attr.Measurement(
        title=u'Radius',
        description=u'The radius of the circle.',
        required=True)

class Circle(CanvasRMLDirective):
    signature = ICircle
    callable = 'circle'
    attrMapping = {'x': 'x_cen', 'y': 'y_cen', 'radius': 'r'}


class IRectangle(IShape):
    """Draws an ellipse on the canvas."""

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the rectangle.',
        required=True)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the rectangle.',
        required=True)

    round = attr.Measurement(
        title=u'Corner Radius',
        description=u'The radius of the rounded corners.',
        required=False)

    href = attr.Text(
        title=u'Link URL',
        description=u'When specified, the rectangle becomes a link to that URL.',
        required=False)

    destination = attr.Text(
        title=u'Link Destination',
        description=(u'When specified, the rectangle becomes a link to that '
                     u'destination.'),
        required=False)

class Rectangle(CanvasRMLDirective):
    signature = IRectangle
    callable = 'rect'
    attrMapping = {'round': 'radius'}

    def process(self):
        if 'round' in self.element.keys():
            self.callable = 'roundRect'
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        # Create a link
        url = kwargs.pop('href', None)
        if url:
            canvas.linkURL(
                url,
                (kwargs['x'], kwargs['y'],
                 kwargs['x']+kwargs['width'], kwargs['y']+kwargs['height']))
        dest = kwargs.pop('destination', None)
        if dest:
            canvas.linkRect(
                '', dest,
                (kwargs['x'], kwargs['y'],
                 kwargs['x']+kwargs['width'], kwargs['y']+kwargs['height']))

        # Render the rectangle
        getattr(canvas, self.callable)(**kwargs)

class IGrid(interfaces.IRMLDirectiveSignature):
    """A shape to be drawn on the canvas."""

    xs = attr.Sequence(
        title=u'X-Coordinates',
        description=(u'A sequence x-coordinates that represent the vertical '
                     u'line positions.'),
        value_type=attr.Measurement(),
        required=True)

    ys = attr.Sequence(
        title=u'Y-Coordinates',
        description=(u'A sequence y-coordinates that represent the horizontal '
                     u'line positions.'),
        value_type=attr.Measurement(),
        required=True)


class Grid(CanvasRMLDirective):
    signature = IGrid
    callable = 'grid'
    attrMapping = {'xs': 'xlist', 'ys': 'ylist'}


class ILines(interfaces.IRMLDirectiveSignature):
    """A path of connected lines drawn on the canvas."""

    linelist = attr.TextNodeGrid(
        title=u'Line List',
        description=(u'A list of lines coordinates to draw.'),
        value_type=attr.Measurement(),
        columns=4,
        required=True)

class Lines(CanvasRMLDirective):
    signature = ILines
    callable = 'lines'


class ICurves(interfaces.IRMLDirectiveSignature):
    """A path of connected bezier curves drawn on the canvas."""

    curvelist = attr.TextNodeGrid(
        title=u'Curve List',
        description=(u'A list of curve coordinates to draw.'),
        value_type=attr.Measurement(),
        columns=8,
        required=True)

class Curves(CanvasRMLDirective):
    signature = ICurves
    callable = 'bezier'

    def process(self):
        argset = self.getAttributeValues(valuesOnly=True)[0]
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        for args in argset:
            getattr(canvas, self.callable)(*args)


class IImage(interfaces.IRMLDirectiveSignature):
    """Draws an external image on the canvas."""

    file = attr.Image(
        title=u'File',
        description=(u'Reference to the external file of the iamge.'),
        required=True)

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'shape.'),
        required=True)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'shape.'),
        required=True)

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the image.',
        required=False)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the image.',
        required=False)

    showBoundary = attr.Boolean(
        title=u'Show Boundary',
        description=(u'A flag determining whether a border should be drawn '
                     u'around the image.'),
        default=False,
        required=False)

    preserveAspectRatio = attr.Boolean(
        title=u'Preserve Aspect Ratio',
        description=(u"A flag determining whether the image's aspect ration "
                     u"should be conserved under any circumstances."),
        default=False,
        required=False)

    mask = attr.Color(
        title=u'Mask',
        description=u'The color mask used to render the image, or "auto" to use the alpha channel if available.',
        default='auto',
        required=False,
        acceptAuto=True)

class Image(CanvasRMLDirective):
    signature = IImage
    callable = 'drawImage'
    attrMapping = {'file': 'image'}

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        preserve = kwargs.pop('preserveAspectRatio')
        show = kwargs.pop('showBoundary')

        if preserve:
            imgX, imgY = kwargs['image'].getSize()

            # Scale image correctly, if width and/or height were specified
            if 'width' in kwargs and 'height' not in kwargs:
                kwargs['height'] = imgY * kwargs['width'] / imgX
            elif 'height' in kwargs and 'width' not in kwargs:
                kwargs['width'] = imgX * kwargs['height'] / imgY
            elif 'width' in kwargs and 'height' in kwargs:
                if float(kwargs['width'])/kwargs['height'] > float(imgX)/imgY:
                    kwargs['width'] = imgX * kwargs['height'] / imgY
                else:
                    kwargs['height'] = imgY * kwargs['width'] / imgX

        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        getattr(canvas, self.callable)(**kwargs)

        if show:
            width = kwargs.get('width', kwargs['image'].getSize()[0])
            height = kwargs.get('height', kwargs['image'].getSize()[1])
            canvas.rect(kwargs['x'], kwargs['y'], width, height)


class IPlace(interfaces.IRMLDirectiveSignature):
    """Draws a set of flowables on the canvas within a given region."""

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'place.'),
        required=True)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'place.'),
        required=True)

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the place.',
        required=False)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the place.',
        required=False)

class Place(CanvasRMLDirective):
    signature = IPlace

    def process(self):
        x, y, width, height = self.getAttributeValues(
            select=('x', 'y', 'width', 'height'), valuesOnly=True)
        y += height

        flows = flowable.Flow(self.element, self.parent)
        flows.process()

        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        for flow in flows.flow:
            flow.canv = canvas
            flowWidth, flowHeight = flow.wrap(width, height)
            if flowWidth <= width and flowHeight <= height:
                y -= flowHeight
                flow.drawOn(canvas, x, y)
                height -= flowHeight
            else:
                raise ValueError("Not enough space")


class IParam(interfaces.IRMLDirectiveSignature):
    """Sets one paramter for the text annotation."""

    name = attr.String(
        title=u'Name',
        description=u'The name of the paramter.',
        required=True)

    value = attr.TextNode(
        title=u'Value',
        description=(u'The parameter value.'),
        required=True)

class Param(directive.RMLDirective):
    signature = IParam

    def process(self):
        args = dict(self.getAttributeValues())
        self.parent.params[args['name']] = args['value']


class ITextAnnotation(interfaces.IRMLDirectiveSignature):
    """Writes a low-level text annotation into the PDF."""
    occurence.containing(
        occurence.ZeroOrMore('param', IParam))

    contents = attr.FirstLevelTextNode(
        title=u'Contents',
        description=u'The PDF commands that are inserted as annotation.',
        required=True)

class TextAnnotation(CanvasRMLDirective):
    signature = ITextAnnotation
    factories = {'param': Param}

    paramTypes = {'escape': attr.Integer()}

    def process(self):
        contents = self.getAttributeValues(valuesOnly=True)[0]
        self.params = {}
        self.processSubDirectives()
        for name, type in self.paramTypes.items():
            if name in self.params:
                bound = type.bind(self)
                self.params[name] = bound.fromUnicode(self.params[name])
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        canvas.textAnnotation(contents, **self.params)


class IMoveTo(interfaces.IRMLDirectiveSignature):
    """Move the path cursor to the specified location."""

    position = attr.TextNodeSequence(
        title=u'Position',
        description=u'Position to which the path pointer is moved to.',
        value_type=attr.Measurement(),
        min_length=2,
        max_length=2,
        required=True)

class MoveTo(directive.RMLDirective):
    signature = IMoveTo

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        self.parent.path.moveTo(*args[0])


class ICurveTo(interfaces.IRMLDirectiveSignature):
    """Create a bezier curve from the current location to the specified one."""

    curvelist = attr.TextNodeGrid(
        title=u'Curve Specification',
        description=u'Describes the end position and the curve properties.',
        value_type=attr.Measurement(),
        columns=6,
        required=True)

class CurveTo(directive.RMLDirective):
    signature = ICurveTo

    def process(self):
        argset = self.getAttributeValues(valuesOnly=True)[0]
        for args in argset:
            self.parent.path.curveTo(*args)

class ICurvesTo(ICurveTo):
    pass
directive.DeprecatedDirective(
    ICurvesTo,
    'Available for ReportLab RML compatibility. Please use the "curveto" '
    'directive instead.')


class IPath(IShape):
    """Create a line path."""
    occurence.containing(
        occurence.ZeroOrMore('moveto', IMoveTo),
        occurence.ZeroOrMore('curveto', ICurveTo),
        occurence.ZeroOrMore('curvesto', ICurvesTo),
        )

    points = attr.TextNodeGrid(
        title=u'Points',
        description=(u'A list of coordinate points that define th path.'),
        value_type=attr.Measurement(),
        columns=2,
        required=True)

    close = attr.Boolean(
        title=u'Close Path',
        description=(u"A flag specifying whether the path should be closed."),
        default=False,
        required=False)

    clip = attr.Boolean(
        title=u'Clip Path',
        description=(u"A flag specifying whether the path should clip "
                     u"overlapping elements."),
        default=False,
        required=False)

class Path(CanvasRMLDirective):
    signature = IPath
    factories = {
        'moveto': MoveTo,
        'curveto': CurveTo,
        'curvesto': CurveTo
        }

    def processPoints(self, text):
        if text.strip() == '':
            return
        bound = self.signature['points'].bind(self)
        for coords in bound.fromUnicode(text):
            self.path.lineTo(*coords)

    def process(self):
        kwargs = dict(self.getAttributeValues(ignore=('points',)))

        # Start the path and set the cursor to the start location.
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        self.path = canvas.beginPath()
        self.path.moveTo(kwargs.pop('x'), kwargs.pop('y'))

        # Process the text before the first sub-directive.
        if self.element.text is not None:
            self.processPoints(self.element.text)
        # Handle each sub-directive.
        for directive in self.element.getchildren():
            if directive.tag in self.factories:
                self.factories[directive.tag](directive, self).process()
            # If there is more text after sub-directive, process it.
            if directive.tail is not None:
                self.processPoints(directive.tail)

        if kwargs.pop('close', False):
            self.path.close()

        if kwargs.pop('clip', False):
            canvas.clipPath(self.path, **kwargs)
        else:
            canvas.drawPath(self.path, **kwargs)


class IFill(interfaces.IRMLDirectiveSignature):
    """Set the fill color."""

    color = attr.Color(
        title=u'Color',
        description=(u'The color value to be set.'),
        required=True)

class Fill(CanvasRMLDirective):
    signature = IFill
    callable = 'setFillColor'
    attrMapping = {'color': 'aColor'}


class IStroke(interfaces.IRMLDirectiveSignature):
    """Set the stroke/line color."""

    color = attr.Color(
        title=u'Color',
        description=(u'The color value to be set.'),
        required=True)

class Stroke(CanvasRMLDirective):
    signature = IStroke
    callable = 'setStrokeColor'
    attrMapping = {'color': 'aColor'}


class ISetFont(interfaces.IRMLDirectiveSignature):
    """Set the font name and/or size."""

    name = attr.String(
        title=u'Font Name',
        description=(u'The name of the font as it was registered.'),
        required=True)

    size = attr.Measurement(
        title=u'Size',
        description=(u'The font size.'),
        required=True)

    leading = attr.Measurement(
        title=u'Leading',
        description=(u'The font leading.'),
        required=False)

class SetFont(CanvasRMLDirective):
    signature = ISetFont
    callable = 'setFont'
    attrMapping = {'name': 'psfontname'}


class ISetFontSize(interfaces.IRMLDirectiveSignature):
    """Set the font size."""

    size = attr.Measurement(
        title=u'Size',
        description=(u'The font size.'),
        required=True)

    leading = attr.Measurement(
        title=u'Leading',
        description=(u'The font leading.'),
        required=False)

class SetFontSize(CanvasRMLDirective):
    signature = ISetFontSize
    callable = 'setFontSize'


class IScale(interfaces.IRMLDirectiveSignature):
    """Scale the drawing using x and y scaling factors."""

    sx = attr.Float(
        title=u'X-Scaling-Factor',
        description=(u'The scaling factor applied on x-coordinates.'),
        default=1,
        required=False)

    sy = attr.Float(
        title=u'Y-Scaling-Factor',
        description=(u'The scaling factor applied on y-coordinates.'),
        default=1,
        required=False)

class Scale(CanvasRMLDirective):
    signature = IScale
    callable = 'scale'
    attrMapping = {'sx': 'x', 'sy': 'y'}


class ITranslate(interfaces.IRMLDirectiveSignature):
    """Translate the drawing coordinates by the specified x and y offset."""

    dx = attr.Measurement(
        title=u'X-Offset',
        description=(u'The amount to move the drawing to the right.'),
        required=True)

    dy = attr.Measurement(
        title=u'Y-Offset',
        description=(u'The amount to move the drawing upward.'),
        required=True)

class Translate(CanvasRMLDirective):
    signature = ITranslate
    callable = 'translate'


class IRotate(interfaces.IRMLDirectiveSignature):
    """Rotate the drawing counterclockwise."""

    degrees = attr.Measurement(
        title=u'Angle',
        description=(u'The angle in degrees.'),
        required=True)

class Rotate(CanvasRMLDirective):
    signature = IRotate
    callable = 'rotate'
    attrMapping = {'degrees': 'theta'}


class ISkew(interfaces.IRMLDirectiveSignature):
    """Skew the drawing."""

    alpha = attr.Measurement(
        title=u'Alpha',
        description=(u'The amount to skew the drawing in the horizontal.'),
        required=True)

    beta = attr.Measurement(
        title=u'Beta',
        description=(u'The amount to skew the drawing in the vertical.'),
        required=True)

class Skew(CanvasRMLDirective):
    signature = ISkew
    callable = 'skew'


class ITransform(interfaces.IRMLDirectiveSignature):
    """A full 2-D matrix transformation"""

    matrix = attr.TextNodeSequence(
        title=u'Matrix',
        description=u'The transformation matrix.',
        value_type=attr.Float(),
        min_length=6,
        max_length=6,
        required=True)

class Transform(CanvasRMLDirective):
    signature = ITransform

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        canvas.transform(*args[0])


class ILineMode(interfaces.IRMLDirectiveSignature):
    """Set the line mode for the following graphics elements."""

    width = attr.Measurement(
        title=u'Width',
        description=(u'The line width.'),
        required=False)

    dash = attr.Sequence(
        title=u'Dash-Pattern',
        description=(u'The dash-pattern of a line.'),
        value_type=attr.Measurement(),
        required=False)

    miterLimit = attr.Measurement(
        title=u'Miter Limit',
        description=(u'The ???.'),
        required=False)

    join = attr.Choice(
        title=u'Join',
        description=u'The way lines are joined together.',
        choices=interfaces.JOIN_CHOICES,
        required=False)

    cap = attr.Choice(
        title=u'Cap',
        description=u'The cap is the desciption of how the line-endings look.',
        choices=interfaces.CAP_CHOICES,
        required=False)

class LineMode(CanvasRMLDirective):
    signature = ILineMode

    def process(self):
        kw = dict(self.getAttributeValues())
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        if 'width' in kw:
            canvas.setLineWidth(kw['width'])
        if 'join' in kw:
            canvas.setLineJoin(kw['join'])
        if 'cap' in kw:
            canvas.setLineCap(kw['cap'])
        if 'miterLimit' in kw:
            canvas.setMiterLimit(kw['miterLimit'])
        if 'dash' in kw:
            canvas.setDash(kw['dash'])


class IBookmark(interfaces.IRMLDirectiveSignature):
    """
    This creates a bookmark to the current page which can be referred to with
    the given key elsewhere. (Used inside a page drawing.)
    """

    name = attr.Text(
        title=u'Name',
        description=u'The name of the bookmark.',
        required=True)

    fit = attr.Choice(
        title=u'Fit',
        description=u'The Fit Type.',
        choices=('XYZ', 'Fit', 'FitH', 'FitV', 'FitR'),
        required=False)

    zoom = attr.Float(
        title=u'Zoom',
        description=u'The zoom level when clicking on the bookmark.',
        required=False)

    x = attr.Measurement(
        title=u'X-Position',
        description=u'The x-position.',
        required=False)

    y = attr.Measurement(
        title=u'Y-Position',
        description=u'The y-position.',
        required=False)

class Bookmark(CanvasRMLDirective):
    signature = IBookmark

    def process(self):
        args = dict(self.getAttributeValues())
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        args['left'], args['top'] = canvas.absolutePosition(args['x'], args['y'])
        canvas.bookmarkPage(**args)


class IPlugInGraphic(interfaces.IRMLDirectiveSignature):
    """Inserts a custom graphic developed in Python."""

    module = attr.String(
        title=u'Module',
        description=u'The Python module in which the flowable is located.',
        required=True)

    function = attr.String(
        title=u'Function',
        description=(u'The name of the factory function within the module '
                     u'that returns the custom flowable.'),
        required=True)

    params = attr.TextNode(
        title=u'Parameters',
        description=(u'A list of parameters encoded as a long string.'),
        required=False)

class PlugInGraphic(CanvasRMLDirective):
    signature = IPlugInGraphic

    def process(self):
        modulePath, functionName, params = self.getAttributeValues(
            valuesOnly=True)
        module = __import__(modulePath, {}, {}, [modulePath])
        function = getattr(module, functionName)
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        function(canvas, params)


class IDrawing(interfaces.IRMLDirectiveSignature):
    """A container directive for all directives that draw directly on the
    cnavas."""
    occurence.containing(
        # State Manipulation
        occurence.ZeroOrMore('saveState', ISaveState),
        occurence.ZeroOrMore('restoreState', IRestoreState),
        # String Drawing
        occurence.ZeroOrMore('drawString', IDrawString),
        occurence.ZeroOrMore('drawRightString', IDrawRightString),
        occurence.ZeroOrMore('drawCenteredString', IDrawCenteredString),
        occurence.ZeroOrMore('drawCentredString', IDrawCenteredString),
        occurence.ZeroOrMore('drawAlignedString', IDrawAlignedString),
        # Drawing Operations
        occurence.ZeroOrMore('ellipse', IEllipse),
        occurence.ZeroOrMore('circle', ICircle),
        occurence.ZeroOrMore('rect', IRectangle),
        occurence.ZeroOrMore('grid', IGrid),
        occurence.ZeroOrMore('lines', ILines),
        occurence.ZeroOrMore('curves', ICurves),
        occurence.ZeroOrMore('image', IImage),
        occurence.ZeroOrMore('place', IPlace),
        occurence.ZeroOrMore('textAnnotation', ITextAnnotation),
        occurence.ZeroOrMore('path', IPath),
        # State Change Operations
        occurence.ZeroOrMore('fill', IFill),
        occurence.ZeroOrMore('stroke', IStroke),
        occurence.ZeroOrMore('setFont', ISetFont),
        occurence.ZeroOrMore('setFontSize', ISetFontSize),
        occurence.ZeroOrMore('scale', IScale),
        occurence.ZeroOrMore('translate', ITranslate),
        occurence.ZeroOrMore('rotate', IRotate),
        occurence.ZeroOrMore('skew', ISkew),
        occurence.ZeroOrMore('transform', ITransform),
        occurence.ZeroOrMore('lineMode', ILineMode),
        # Form Field Elements
        occurence.ZeroOrMore('barCode', form.IBarCode),
        occurence.ZeroOrMore('textField', form.ITextField),
        occurence.ZeroOrMore('buttonField', form.IButtonField),
        occurence.ZeroOrMore('selectField', form.ISelectField),
        # Charts
        occurence.ZeroOrMore('barChart', chart.IBarChart),
        occurence.ZeroOrMore('barChart3D', chart.IBarChart3D),
        occurence.ZeroOrMore('linePlot', chart.ILinePlot),
        occurence.ZeroOrMore('linePlot3D', chart.ILinePlot3D),
        occurence.ZeroOrMore('pieChart', chart.IPieChart),
        occurence.ZeroOrMore('pieChart3D', chart.IPieChart3D),
        occurence.ZeroOrMore('spiderChart', chart.ISpiderChart),
        # Misc
        occurence.ZeroOrMore('bookmark', IBookmark),
        occurence.ZeroOrMore('plugInGraphic', IPlugInGraphic),
        )

class Drawing(directive.RMLDirective):
    signature = IDrawing

    factories = {
        # State Management
        'saveState': SaveState,
        'restoreState': RestoreState,
        # Drawing Strings
        'drawString': DrawString,
        'drawRightString': DrawRightString,
        'drawCenteredString': DrawCenteredString,
        'drawCentredString': DrawCenteredString,
        'drawAlignedString': DrawAlignedString,
        # Drawing Operations
        'ellipse': Ellipse,
        'circle': Circle,
        'rect': Rectangle,
        'grid': Grid,
        'lines': Lines,
        'curves': Curves,
        'image': Image,
        'place': Place,
        'textAnnotation': TextAnnotation,
        'path': Path,
        # Form Field Elements
        'barCode': form.BarCode,
        'textField': form.TextField,
        'buttonField': form.ButtonField,
        'selectField': form.SelectField,
        # State Change Operations
        'fill': Fill,
        'stroke': Stroke,
        'setFont': SetFont,
        'setFontSize': SetFontSize,
        'scale': Scale,
        'translate': Translate,
        'rotate': Rotate,
        'skew': Skew,
        'transform': Transform,
        'lineMode': LineMode,
        # Charts
        'barChart': chart.BarChart,
        'barChart3D': chart.BarChart3D,
        'linePlot': chart.LinePlot,
        'linePlot3D': chart.LinePlot3D,
        'pieChart': chart.PieChart,
        'pieChart3D': chart.PieChart3D,
        'spiderChart': chart.SpiderChart,
        # Misc
        'bookmark': Bookmark,
        'plugInGraphic': PlugInGraphic,
        }


class IPageDrawing(IDrawing):
    """Draws directly on the content of one page's canvas. Every call of this
    directive creates a new page."""

    occurence.containing(
        #'mergePage': IMergePage,
        *IDrawing.getTaggedValue('directives'))

class PageDrawing(Drawing):
    signature = IDrawing

    factories = Drawing.factories.copy()
    factories.update({
        'mergePage': page.MergePage
        })

    def process(self):
        super(Drawing, self).process()
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        canvas.showPage()


class IPageInfo(interfaces.IRMLDirectiveSignature):
    """Set's up page-global settings."""

    pageSize = attr.PageSize(
        title=u'Page Size',
        description=(u'The page size of all pages within this document.'),
        required=True)

class PageInfo(CanvasRMLDirective):
    signature=IPageInfo
    callable = 'setPageSize'
    attrMapping = {'pageSize': 'size'}



# Hook up our custom paragraph parser.
import z3c.rml.paraparser
import z3c.rml.rlfix

import pkg_resources
__version__ = pkg_resources.require('z3c.rml')[0].version

from reportlab.lib.styles import getSampleStyleSheet
SampleStyleSheet = getSampleStyleSheet()



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Test generating reference.
"""
import os
import unittest
from z3c.rml import reference


class ReferenceTestCase(unittest.TestCase):

    level = 3

    def runTest(self):
        reference.main(
            os.path.join(os.path.dirname(reference.__file__),
                         'rml-reference.pdf'))


def test_suite():
    return unittest.TestSuite([ReferenceTestCase()])



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Chart Element Processing
"""
import reportlab.lib.formatters
from reportlab.graphics import shapes
from reportlab.graphics.charts import barcharts, lineplots, piecharts
from reportlab.graphics.charts import spider, doughnut
from z3c.rml import attr, directive, interfaces, occurence

# Patches against Reportlab 2.0
lineplots.Formatter = reportlab.lib.formatters.Formatter


class PropertyItem(directive.RMLDirective):

    def process(self):
        attrs = dict(self.getAttributeValues())
        self.parent.dataList.append(attrs)


class PropertyCollection(directive.RMLDirective):
    propertyName = None

    def processAttributes(self):
        prop = getattr(self.parent.context, self.propertyName)
        # Get global properties
        for name, value in self.getAttributeValues():
            setattr(prop, name, value)

    def process(self):
        self.processAttributes()
        # Get item specific properties
        prop = getattr(self.parent.context, self.propertyName)
        self.dataList = []
        self.processSubDirectives()
        for index, data in enumerate(self.dataList):
            for name, value in data.items():
                setattr(prop[index], name, value)


class IText(interfaces.IRMLDirectiveSignature):
    """Draw a text on the chart."""

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'text.'),
        required=True)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'text.'),
        required=True)

    angle = attr.Float(
        title=u'Rotation Angle',
        description=(u'The angle about which the text will be rotated.'),
        required=False)

    text = attr.TextNode(
        title=u'Text',
        description=u'The text to be printed.',
        required=True)

    fontName = attr.String(
        title=u'Font Name',
        description=u'The name of the font.',
        required=False)

    fontSize = attr.Measurement(
        title=u'Font Size',
        description=u'The font size for the text.',
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=u'The color in which the text will appear.',
        required=False)

    textAnchor = attr.Choice(
        title=u'Text Anchor',
        description=u'The position in the text to which the coordinates refer.',
        choices=('start', 'middle', 'end', 'boxauto'),
        required=False)


class Text(directive.RMLDirective):
    signature = IText

    def process(self):
        attrs = dict(self.getAttributeValues())
        string = shapes.String(
            attrs.pop('x'), attrs.pop('y'), attrs.pop('text'))
        angle = attrs.pop('angle', 0)
        for name, value in attrs.items():
            setattr(string, name, value)
        group = shapes.Group(string)
        group.translate(0,0)
        group.rotate(angle)
        self.parent.parent.drawing.add(group)


class ITexts(interfaces.IRMLDirectiveSignature):
    """A set of texts drawn on the chart."""
    occurence.containing(
        occurence.ZeroOrMore('text', IText)
        )

class Texts(directive.RMLDirective):
    signature = ITexts
    factories = {'text': Text}


class Series(directive.RMLDirective):

    def process(self):
        attrs = self.getAttributeValues(valuesOnly=True)
        self.parent.data.append(attrs[0])


class Data(directive.RMLDirective):
    series = None

    def process(self):
        self.data = []
        self.factories = {'series': self.series}
        self.processSubDirectives()
        self.parent.context.data = self.data


class ISeries1D(interfaces.IRMLDirectiveSignature):
    """A one-dimensional series."""

    values = attr.TextNodeSequence(
        title=u'Values',
        description=u"Numerical values representing the series' data.",
        value_type=attr.Float(),
        required=True)

class Series1D(Series):
    signature = ISeries1D


class IData1D(interfaces.IRMLDirectiveSignature):
    """A 1-D data set."""
    occurence.containing(
        occurence.OneOrMore('series', ISeries1D)
        )

class Data1D(Data):
    signature = IData1D
    series = Series1D


class ISingleData1D(interfaces.IRMLDirectiveSignature):
    """A 1-D data set."""
    occurence.containing(
        occurence.One('series', ISeries1D)
        )

class SingleData1D(Data1D):
    signature = ISingleData1D

    def process(self):
        self.data = []
        self.factories = {'series': self.series}
        self.processSubDirectives()
        self.parent.context.data = self.data[0]


class ISeries2D(interfaces.IRMLDirectiveSignature):
    """A two-dimensional series."""

    values = attr.TextNodeGrid(
        title=u'Values',
        description=u"Numerical values representing the series' data.",
        value_type=attr.Float(),
        columns=2,
        required=True)

class Series2D(Series):
    signature = ISeries2D


class IData2D(interfaces.IRMLDirectiveSignature):
    """A 2-D data set."""
    occurence.containing(
        occurence.OneOrMore('series', ISeries2D)
        )

class Data2D(Data):
    signature = IData2D
    series = Series2D


class IBar(interfaces.IRMLDirectiveSignature):
    """Define the look of a bar."""

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=u'The color in which the bar border is drawn.',
        required=False)

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The width of the bar border line.',
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=u'The color with which the bar is filled.',
        required=False)

class Bar(PropertyItem):
    signature = IBar

class IBars(IBar):
    """Collection of bar subscriptions."""
    occurence.containing(
        occurence.ZeroOrMore('bar', IBar)
        )

class Bars(PropertyCollection):
    signature = IBars
    propertyName = 'bars'
    factories = {'bar': Bar}


class ILabelBase(interfaces.IRMLDirectiveSignature):

    dx = attr.Measurement(
        title=u'Horizontal Extension',
        description=(u'The width of the label.'),
        required=False)

    dy = attr.Measurement(
        title=u'Vertical Extension',
        description=(u'The height of the label.'),
        required=False)

    angle = attr.Float(
        title=u'Angle',
        description=(u'The angle to rotate the label.'),
        required=False)

    boxAnchor = attr.Choice(
        title=u'Box Anchor',
        description=(u'The position relative to the label.'),
        choices=('nw','n','ne','w','c','e','sw','s','se', 'autox', 'autoy'),
        required=False)

    boxStrokeColor = attr.Color(
        title=u'Box Stroke Color',
        description=(u'The color of the box border line.'),
        required=False)

    boxStrokeWidth = attr.Measurement(
        title=u'Box Stroke Width',
        description=u'The width of the box border line.',
        required=False)

    boxFillColor = attr.Color(
        title=u'Box Fill Color',
        description=(u'The color in which the box is filled.'),
        required=False)

    boxTarget = attr.Text(
        title=u'Box Target',
        description=u'The box target.',
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=(u'The color in which the label is filled.'),
        required=False)

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=(u'The color of the label.'),
        required=False)

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The width of the label line.',
        required=False)

    fontName = attr.String(
        title=u'Font Name',
        description=u'The font used to print the value.',
        required=False)

    fontSize = attr.Measurement(
        title=u'Font Size',
        description=u'The size of the value text.',
        required=False)

    leading = attr.Measurement(
        title=u'Leading',
        description=(u'The height of a single text line. It includes '
                     u'character height.'),
        required=False)

    width = attr.Measurement(
        title=u'Width',
        description=u'The width the label.',
        required=False)

    maxWidth = attr.Measurement(
        title=u'Maximum Width',
        description=u'The maximum width the label.',
        required=False)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height the label.',
        required=False)

    textAnchor = attr.Choice(
        title=u'Text Anchor',
        description=u'The position in the text to which the coordinates refer.',
        choices=('start', 'middle', 'end', 'boxauto'),
        required=False)

    visible = attr.Boolean(
        title=u'Visible',
        description=u'A flag making the label text visible.',
        required=False)

    leftPadding = attr.Measurement(
        title=u'Left Padding',
        description=u'The size of the padding on the left side.',
        required=False)

    rightPadding = attr.Measurement(
        title=u'Right Padding',
        description=u'The size of the padding on the right side.',
        required=False)

    topPadding = attr.Measurement(
        title=u'Top Padding',
        description=u'The size of the padding on the top.',
        required=False)

    bottomPadding = attr.Measurement(
        title=u'Bottom Padding',
        description=u'The size of the padding on the bottom.',
        required=False)


class IPositionLabelBase(ILabelBase):

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'label.'),
        required=False)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'label.'),
        required=False)


class ILabel(IPositionLabelBase):
    """A label for the chart on an axis."""

    text = attr.TextNode(
        title=u'Text',
        description=u'The label text to be displayed.',
        required=True)

class Label(PropertyItem):
    signature = ILabel

class IBarLabels(ILabelBase):
    """A set of labels for a bar chart"""
    occurence.containing(
        occurence.ZeroOrMore('label', ILabel)
        )

class BarLabels(PropertyCollection):
    signature = IBarLabels
    propertyName = 'barLabels'
    factories = {'label': Label}
    name = 'barLabels'

class ILabels(IPositionLabelBase):
    """A set of labels of an axis."""
    occurence.containing(
        occurence.ZeroOrMore('label', ILabel)
        )

class Labels(PropertyCollection):
    signature = ILabels
    propertyName = 'labels'
    factories = {'label': Label}


class IAxis(interfaces.IRMLDirectiveSignature):
    occurence.containing(
        occurence.ZeroOrMore('labels', ILabels)
        )

    visible = attr.Boolean(
        title=u'Visible',
        description=u'When true, draw the entire axis with all details.',
        required=False)

    visibleAxis = attr.Boolean(
        title=u'Visible Axis',
        description=u'When true, draw the axis line.',
        required=False)

    visibleTicks = attr.Boolean(
        title=u'Visible Ticks',
        description=u'When true, draw the axis ticks on the line.',
        required=False)

    visibleLabels = attr.Boolean(
        title=u'Visible Labels',
        description=u'When true, draw the axis labels.',
        required=False)

    visibleGrid = attr.Boolean(
        title=u'Visible Grid',
        description=u'When true, draw the grid lines for the axis.',
        required=False)

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The width of axis line and ticks.',
        required=False)

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=u'The color in which the axis line and ticks are drawn.',
        required=False)

    strokeDashArray = attr.Sequence(
        title=u'Stroke Dash Array',
        description=u'The dash array that is used for the axis line and ticks.',
        value_type=attr.Float(),
        required=False)

    gridStrokeWidth = attr.Measurement(
        title=u'Grid Stroke Width',
        description=u'The width of the grid lines.',
        required=False)

    gridStrokeColor = attr.Color(
        title=u'Grid Stroke Color',
        description=u'The color in which the grid lines are drawn.',
        required=False)

    gridStrokeDashArray = attr.Sequence(
        title=u'Grid Stroke Dash Array',
        description=u'The dash array that is used for the grid lines.',
        value_type=attr.Float(),
        required=False)

    gridStart = attr.Measurement(
        title=u'Grid Start',
        description=(u'The start of the grid lines with respect to the '
                     u'axis origin.'),
        required=False)

    gridEnd = attr.Measurement(
        title=u'Grid End',
        description=(u'The end of the grid lines with respect to the '
                     u'axis origin.'),
        required=False)

    style = attr.Choice(
        title=u'Style',
        description=u'The plot style of the common categories.',
        choices=('parallel', 'stacked', 'parallel_3d'),
        required=False)


class Axis(directive.RMLDirective):
    signature = IAxis
    name = ''
    factories = {'labels': Labels}

    def process(self):
        self.context = axis = getattr(self.parent.context, self.name)
        for name, value in self.getAttributeValues():
            setattr(axis, name, value)
        self.processSubDirectives()


class IName(interfaces.IRMLDirectiveSignature):
    """A category name"""
    text = attr.TextNode(
        title=u'Text',
        description=u'The text value that is the name.',
        required=True)

class Name(directive.RMLDirective):
    signature = IName

    def process(self):
        text = self.getAttributeValues(valuesOnly=True)[0]
        self.parent.names.append(text)


class ICategoryNames(interfaces.IRMLDirectiveSignature):
    """A list of category names."""
    occurence.containing(
        occurence.OneOrMore('name', IName),
        )

class CategoryNames(directive.RMLDirective):
    signature = ICategoryNames
    factories = {'name': Name}

    def process(self):
        self.names = []
        self.processSubDirectives()
        self.parent.context.categoryNames = self.names


class ICategoryAxis(IAxis):
    """An axis displaying categories (instead of numerical values)."""
    occurence.containing(
        occurence.ZeroOrOne('categoryNames', ICategoryNames),
        *IAxis.queryTaggedValue('directives', ())
        )

    categoryNames = attr.Sequence(
        title=u'Category Names',
        description=u'A simple list of category names.',
        value_type=attr.Text(),
        required=False)

    joinAxis = attr.Boolean(
        title=u'Join Axis',
        description=u'When true, both axes join together.',
        required=False)

    joinAxisPos = attr.Measurement(
        title=u'Join Axis Position',
        description=u'The position at which the axes should join together.',
        required=False)

    reverseDirection = attr.Boolean(
        title=u'Reverse Direction',
        description=u'A flag to reverse the direction of category names.',
        required=False)

    labelAxisMode = attr.Choice(
        title=u'Label Axis Mode',
        description=u'Defines the relative position of the axis labels.',
        choices=('high', 'low', 'axis'),
        required=False)

    tickShift = attr.Boolean(
        title=u'Tick Shift',
        description=(u'When true, place the ticks in the center of a '
                     u'category instead the beginning and end.'),
        required=False)

class CategoryAxis(Axis):
    signature = ICategoryAxis
    name = 'categoryAxis'
    factories = Axis.factories.copy()
    factories.update({
        'categoryNames': CategoryNames,
        })


class IXCategoryAxis(ICategoryAxis):
    """X-Category Axis"""

    tickUp = attr.Measurement(
        title=u'Tick Up',
        description=u'Length of tick above the axis line.',
        required=False)

    tickDown = attr.Measurement(
        title=u'Tick Down',
        description=u'Length of tick below the axis line.',
        required=False)

    joinAxisMode = attr.Choice(
        title=u'Join Axis Mode',
        description=u'Mode for connecting axes.',
        choices=('bottom', 'top', 'value', 'points', 'None'),
        required=False)

class XCategoryAxis(CategoryAxis):
    signature = IXCategoryAxis


class IYCategoryAxis(ICategoryAxis):
    """Y-Category Axis"""

    tickLeft = attr.Measurement(
        title=u'Tick Left',
        description=u'Length of tick left to the axis line.',
        required=False)

    tickRight = attr.Measurement(
        title=u'Tick Right',
        description=u'Length of tick right to the axis line.',
        required=False)

    joinAxisMode = attr.Choice(
        title=u'Join Axis Mode',
        description=u'Mode for connecting axes.',
        choices=('bottom', 'top', 'value', 'points', 'None'),
        required=False)

class YCategoryAxis(CategoryAxis):
    signature = IYCategoryAxis


class IValueAxis(IAxis):

    forceZero = attr.Boolean(
        title=u'Force Zero',
        description=u'When set, the range will contain the origin.',
        required=False)

    minimumTickSpacing = attr.Measurement(
        title=u'Minimum Tick Spacing',
        description=u'The minimum distance between ticks.',
        required=False)

    maximumTicks = attr.Integer(
        title=u'Maximum Ticks',
        description=u'The maximum number of ticks to be shown.',
        required=False)

    labelTextFormat = attr.String(
        title=u'Label Text Format',
        description=u'Formatting string for axis labels.',
        required=False)

    labelTextPostFormat = attr.Text(
        title=u'Label Text Post Format',
        description=u'An additional formatting string.',
        required=False)

    labelTextScale = attr.Float(
        title=u'Label Text Scale',
        description=u'The sclaing factor for the label tick values.',
        required=False)

    valueMin = attr.Float(
        title=u'Minimum Value',
        description=u'The smallest value on the axis.',
        required=False)

    valueMax = attr.Float(
        title=u'Maximum Value',
        description=u'The largest value on the axis.',
        required=False)

    valueStep = attr.Float(
        title=u'Value Step',
        description=u'The step size between ticks',
        required=False)

    valueSteps = attr.Sequence(
        title=u'Step Sizes',
        description=u'List of step sizes between ticks.',
        value_type = attr.Float(),
        required=False)

    rangeRound = attr.Choice(
        title=u'Range Round',
        description=u'Method to be used to round the range values.',
        choices=('none', 'both', 'ceiling', 'floor'),
        required=False)

    zrangePref = attr.Float(
        title=u'Zero Range Preference',
        description=u'Zero range axis limit preference.',
        required=False)

class ValueAxis(Axis):
    signature = IValueAxis
    name = 'valueAxis'


class IXValueAxis(IValueAxis):
    """X-Value Axis"""

    tickUp = attr.Measurement(
        title=u'Tick Up',
        description=u'Length of tick above the axis line.',
        required=False)

    tickDown = attr.Measurement(
        title=u'Tick Down',
        description=u'Length of tick below the axis line.',
        required=False)

    joinAxis = attr.Boolean(
        title=u'Join Axis',
        description=u'Whether to join the axes.',
        required=False)

    joinAxisMode = attr.Choice(
        title=u'Join Axis Mode',
        description=u'Mode for connecting axes.',
        choices=('bottom', 'top', 'value', 'points', 'None'),
        required=False)

    joinAxisPos = attr.Measurement(
        title=u'Join Axis Position',
        description=u'The position in the plot at which to join the axes.',
        required=False)

class XValueAxis(ValueAxis):
    signature = IXValueAxis

class LineXValueAxis(XValueAxis):
    name = 'xValueAxis'

class IYValueAxis(IValueAxis):
    """Y-Value Axis"""

    tickLeft = attr.Measurement(
        title=u'Tick Left',
        description=u'Length of tick left to the axis line.',
        required=False)

    tickRight = attr.Measurement(
        title=u'Tick Right',
        description=u'Length of tick right to the axis line.',
        required=False)

    joinAxis = attr.Boolean(
        title=u'Join Axis',
        description=u'Whether to join the axes.',
        required=False)

    joinAxisMode = attr.Choice(
        title=u'Join Axis Mode',
        description=u'Mode for connecting axes.',
        choices=('bottom', 'top', 'value', 'points', 'None'),
        required=False)

    joinAxisPos = attr.Measurement(
        title=u'Join Axis Position',
        description=u'The position in the plot at which to join the axes.',
        required=False)


class YValueAxis(ValueAxis):
    signature = IYValueAxis

class LineYValueAxis(YValueAxis):
    name = 'yValueAxis'


class ILineLabels(IPositionLabelBase):
    """A set of labels of an axis."""
    occurence.containing(
        occurence.ZeroOrMore('label', ILabel)
        )

class LineLabels(PropertyCollection):
    signature = ILineLabels
    propertyName = 'lineLabels'
    factories = {'label': Label}


class ILineBase(interfaces.IRMLDirectiveSignature):

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The width of the plot line.',
        required=False)

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=u'The color of the plot line.',
        required=False)

    strokeDashArray = attr.Sequence(
        title=u'Stroke Dash Array',
        description=u'The dash array of the plot line.',
        value_type = attr.Float(),
        required=False)

    symbol = attr.Symbol(
        title=u'Symbol',
        description=u'The symbol to be used for every data point in the plot.',
        required=False)

class ILine(ILineBase):
    """A line description of a series of a line plot."""

    name = attr.Text(
        title=u'Name',
        description=u'The name of the line.',
        required=False)

class Line(PropertyItem):
    signature = ILine

class ILines(ILineBase):
    """The set of all line descriptions in the line plot."""
    occurence.containing(
        occurence.OneOrMore('line', ILine),
        )

class Lines(PropertyCollection):
    signature = ILines
    propertyName = 'lines'
    factories = {'line': Line}


class ISliceLabel(ILabelBase):
    """The label of a slice within a bar chart."""

    text = attr.TextNode(
        title=u'Text',
        description=u'The label text to be displayed.',
        required=True)

class SliceLabel(Label):
    signature = ISliceLabel

    def process(self):
        for name, value in self.getAttributeValues():
            self.parent.context['label_'+name] = value
        # Now we do not have simple labels anymore
        self.parent.parent.parent.context.simpleLabels = False


class ISlicePointer(interfaces.IRMLDirectiveSignature):
    """A pointer to a slice in a pie chart."""

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=u'The color of the pointer line.',
        required=False)

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The wodth of the pointer line.',
        required=False)

    elbowLength = attr.Measurement(
        title=u'Elbow Length',
        description=u'The length of the final segment of the pointer.',
        required=False)

    edgePad = attr.Measurement(
        title=u'Edge Padding',
        description=u'The padding between between the pointer label and box.',
        required=False)

    piePad = attr.Measurement(
        title=u'Pie Padding',
        description=u'The padding between between the pointer label and chart.',
        required=False)

class SlicePointer(directive.RMLDirective):
    signature = ISlicePointer

    def process(self):
        for name, value in self.getAttributeValues():
            self.parent.context['label_pointer_'+name] = value


class ISliceBase(interfaces.IRMLDirectiveSignature):

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The wodth of the slice line.',
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=u'The fill color of the slice.',
        required=False)

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=u'The color of the pointer line.',
        required=False)

    strokeDashArray = attr.Sequence(
        title=u'Stroke Dash Array',
        description=u'Teh dash array of the slice borderline.',
        value_type=attr.Float(),
        required=False)

    popout = attr.Measurement(
        title=u'Popout',
        description=u'The distance of how much the slice should be popped out.',
        required=False)

    fontName = attr.String(
        title=u'Font Name',
        description=u'The font name of the label.',
        required=False)

    fontSize = attr.Measurement(
        title=u'Font Size',
        description=u'The font size of the label.',
        required=False)

    labelRadius = attr.Measurement(
        title=u'Label Radius',
        description=(u'The radius at which the label should be placed around '
                     u'the pie.'),
        required=False)


class ISlice(ISliceBase):
    """A slice in a pie chart."""
    occurence.containing(
        occurence.ZeroOrOne('label', ISliceLabel),
        occurence.ZeroOrOne('pointer', ISlicePointer),
        )

    swatchMarker = attr.Symbol(
        required=False)

class Slice(directive.RMLDirective):
    signature = ISlice
    factories = {
        'label': SliceLabel,
        'pointer': SlicePointer}

    def process(self):
        self.context = attrs = dict(self.getAttributeValues())
        self.processSubDirectives()
        self.parent.context.append(attrs)


class ISlice3D(ISlice):
    """A 3-D slice of a 3-D pie chart."""

    fillColorShaded = attr.Color(
        title=u'Fill Color Shade',
        description=u'The shade used for the fill color.',
        required=False)

class Slice3D(Slice):
    signature = ISlice3D
    factories = {}
    # Sigh, the 3-D Pie does not support advanced slice labels. :-(
    #     'label': SliceLabel}


class ISlices(ISliceBase):
    """The collection of all 2-D slice descriptions."""
    occurence.containing(
        occurence.OneOrMore('slice', ISlice),
        )

class Slices(directive.RMLDirective):
    signature = ISlices
    factories = {'slice': Slice}

    def process(self):
        # Get global slice properties
        for name, value in self.getAttributeValues():
            setattr(self.parent.context.slices, name, value)
        # Get slice specific properties
        self.context = slicesData = []
        self.processSubDirectives()
        for index, sliceData in enumerate(slicesData):
            for name, value in sliceData.items():
                setattr(self.parent.context.slices[index], name, value)


class ISlices3D(ISliceBase):
    """The collection of all 3-D slice descriptions."""
    occurence.containing(
        occurence.OneOrMore('slice', ISlice3D),
        )

    fillColorShaded = attr.Color(
        required=False)

class Slices3D(Slices):
    signature = ISlices3D
    factories = {'slice': Slice3D}


class ISimpleLabel(IName):
    """A simple label"""

class SimpleLabel(Name):
    signature = ISimpleLabel

class ISimpleLabels(interfaces.IRMLDirectiveSignature):
    """A set of simple labels for a chart."""
    occurence.containing(
        occurence.OneOrMore('label', ISimpleLabel),
        )

class SimpleLabels(directive.RMLDirective):
    signature = ISimpleLabels
    factories = {'label': Name}

    def process(self):
        self.names = []
        self.processSubDirectives()
        self.parent.context.labels = self.names


class IStrandBase(interfaces.IRMLDirectiveSignature):

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'The line width of the strand.',
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=u'The fill color of the strand area.',
        required=False)

    strokeColor= attr.Color(
        title=u'Stroke Color',
        description=u'The color of the strand line.',
        required=False)

    strokeDashArray = attr.Sequence(
        title=u'Stroke Dash Array',
        description=u'The dash array of the strand line.',
        value_type=attr.Float(),
        required=False)

    symbol = attr.Symbol(
        title=u'Symbol',
        description=u'The symbol to use to mark the strand.',
        required=False)

    symbolSize = attr.Measurement(
        title=u'Symbol Size',
        description=u'The size of the strand symbol.',
        required=False)

class IStrand(IStrandBase):
    """A strand in the spider diagram"""

    name = attr.Text(
        title=u'Name',
        description=u'The name of the strand.',
        required=False)

class Strand(PropertyItem):
    signature = IStrand


class IStrands(IStrand):
    """A collection of strands."""
    occurence.containing(
        occurence.OneOrMore('strand', IStrand)
        )

class Strands(PropertyCollection):
    signature = IStrands
    propertyName = 'strands'
    attrs = IStrandBase
    factories = {'strand': Strand}


class IStrandLabelBase(ILabelBase):

    row = attr.Integer(
        title=u'Row',
        description=u'The row of the strand label',
        required=False)

    col = attr.Integer(
        title=u'Column',
        description=u'The column of the strand label.',
        required=False)

    format = attr.String(
        title=u'Format',
        description=u'The format string for the label.',
        required=False)

class IStrandLabel(IStrandLabelBase):
    """A label for a strand."""

    _text = attr.TextNode(
        title=u'Text',
        description=u'The label text of the strand.',
        required=False)

    dR = attr.Float(
        title=u'Radial Shift',
        description=u'The radial shift of the label.',
        required=False)

class StrandLabel(Label):
    signature = IStrandLabel


class IStrandLabels(IStrandLabelBase):
    """A set of strand labels."""
    occurence.containing(
        occurence.OneOrMore('label', IStrandLabel)
        )

class StrandLabels(PropertyCollection):
    signature = IStrandLabels
    propertyName = 'strandLabels'
    factories = {'label': StrandLabel}

    def process(self):
        self.processAttributes()
        # Get item specific properties
        prop = getattr(self.parent.context, self.propertyName)
        self.dataList = []
        self.processSubDirectives()
        for data in self.dataList:
            row = data.pop('row')
            col = data.pop('col')
            for name, value in data.items():
                setattr(prop[row, col], name, value)


class ISpoke(interfaces.IRMLDirectiveSignature):
    """A spoke in the spider diagram."""

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u"The width of the spoke's line.",
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=u"The fill color of the spoke's area.",
        required=False)

    strokeColor= attr.Color(
        title=u'Stroke Color',
        description=u'The color of the spoke line.',
        required=False)

    strokeDashArray = attr.Sequence(
        title=u'Stroke Dash Array',
        description=u'The dash array of the spoke line.',
        value_type=attr.Float(),
        required=False)

    labelRadius = attr.Measurement(
        title=u'Label Radius',
        description=u'The radius of the label arouns the spoke.',
        required=False)

    visible = attr.Boolean(
        title=u'Visible',
        description=u'When true, the spoke line is drawn.',
        required=False)

class Spoke(PropertyItem):
    signature = ISpoke


class ISpokes(ISpoke):
    """A collection of spokes."""
    occurence.containing(
        occurence.OneOrMore('spoke', ISpoke)
        )

class Spokes(PropertyCollection):
    signature = ISpokes
    propertyName = 'spokes'
    factories = {'spoke': Spoke}


class ISpokeLabelBase(ILabelBase):
    pass

class ISpokeLabel(ISpokeLabelBase):
    """A label for a spoke."""
    _text = attr.TextNode(
        title=u'Text',
        description=u'The text of the spoke (label).',
        required=False)

class SpokeLabel(Label):
    signature = ISpokeLabel


class ISpokeLabels(ISpokeLabelBase):
    """A set of spoke labels."""
    occurence.containing(
        occurence.OneOrMore('label', ISpokeLabel)
        )

class SpokeLabels(PropertyCollection):
    signature = ISpokeLabels
    propertyName = 'spokeLabels'
    factories = {'label': SpokeLabel}


class IChart(interfaces.IRMLDirectiveSignature):
    occurence.containing(
        occurence.ZeroOrOne('texts', ITexts),
        )

    # Drawing Options

    dx = attr.Measurement(
        title=u'Drawing X-Position',
        description=u'The x-position of the entire drawing on the canvas.',
        required=False)

    dy = attr.Measurement(
        title=u'Drawing Y-Position',
        description=u'The y-position of the entire drawing on the canvas.',
        required=False)

    dwidth = attr.Measurement(
        title=u'Drawing Width',
        description=u'The width of the entire drawing',
        required=False)

    dheight = attr.Measurement(
        title=u'Drawing Height',
        description=u'The height of the entire drawing',
        required=False)

    angle = attr.Float(
        title=u'Angle',
        description=u'The orientation of the drawing as an angle in degrees.',
        required=False)

    # Plot Area Options

    x = attr.Measurement(
        title=u'Chart X-Position',
        description=u'The x-position of the chart within the drawing.',
        required=False)

    y = attr.Measurement(
        title=u'Chart Y-Position',
        description=u'The y-position of the chart within the drawing.',
        required=False)

    width = attr.Measurement(
        title=u'Chart Width',
        description=u'The width of the chart.',
        required=False)

    height = attr.Measurement(
        title=u'Chart Height',
        description=u'The height of the chart.',
        required=False)

    strokeColor = attr.Color(
        title=u'Stroke Color',
        description=u'Color of the chart border.',
        required=False)

    strokeWidth = attr.Measurement(
        title=u'Stroke Width',
        description=u'Width of the chart border.',
        required=False)

    fillColor = attr.Color(
        title=u'Fill Color',
        description=u'Color of the chart interior.',
        required=False)

    debug = attr.Boolean(
        title=u'Debugging',
        description=u'A flag that when set to True turns on debug messages.',
        required=False)

class Chart(directive.RMLDirective):
    signature = IChart
    factories = {
        'texts': Texts
        }
    attrMapping = {}

    def createChart(self, attributes):
        raise NotImplementedError

    def process(self):
        attrs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        angle = attrs.pop('angle', 0)
        x, y = attrs.pop('dx'), attrs.pop('dy')
        self.drawing = shapes.Drawing(attrs.pop('dwidth'), attrs.pop('dheight'))
        self.context = chart = self.createChart(attrs)
        self.processSubDirectives()
        group = shapes.Group(chart)
        group.translate(0,0)
        group.rotate(angle)
        self.drawing.add(group)
        manager = attr.getManager(self, interfaces.ICanvasManager)
        self.drawing.drawOn(manager.canvas, x, y)


class IBarChart(IChart):
    """Creates a two-dimensional bar chart."""
    occurence.containing(
        occurence.One('data', IData1D),
        occurence.ZeroOrOne('bars', IBars),
        occurence.ZeroOrOne('categoryAxis', ICategoryAxis),
        occurence.ZeroOrOne('valueAxis', IValueAxis),
        occurence.ZeroOrOne('barLabels', IBarLabels),
        *IChart.queryTaggedValue('directives', ())
        )

    direction = attr.Choice(
        title=u'Direction',
        description=u'The direction of the bars within the chart.',
        choices=('horizontal', 'vertical'),
        default='horizontal',
        required=False)

    useAbsolute = attr.Boolean(
        title=u'Use Absolute Spacing',
        description=u'Flag to use absolute spacing values.',
        default=False,
        required=False)

    barWidth = attr.Measurement(
        title=u'Bar Width',
        description=u'The width of an individual bar.',
        default=10,
        required=False)

    groupSpacing = attr.Measurement(
        title=u'Group Spacing',
        description=u'Width between groups of bars.',
        default=5,
        required=False)

    barSpacing = attr.Measurement(
        title=u'Bar Spacing',
        description=u'Width between individual bars.',
        default=0,
        required=False)

    barLabelFormat = attr.String(
        title=u'Bar Label Text Format',
        description=u'Formatting string for bar labels.',
        required=False)


class BarChart(Chart):
    signature = IBarChart
    nameBase = 'BarChart'
    factories = Chart.factories.copy()
    factories.update({
        'data': Data1D,
        'bars': Bars,
	'barLabels': BarLabels,
        })

    def createChart(self, attrs):
        direction = attrs.pop('direction')
        # Setup sub-elements based on direction
        if direction == 'horizontal':
            self.factories['categoryAxis'] = YCategoryAxis
            self.factories['valueAxis'] = XValueAxis
        else:
            self.factories['categoryAxis'] = XCategoryAxis
            self.factories['valueAxis'] = YValueAxis
        # Generate the chart
        chart = getattr(
            barcharts, direction.capitalize()+self.nameBase)()
        for name, value in attrs.items():
            setattr(chart, name, value)
        return chart


class IBarChart3D(IBarChart):
    """Creates a three-dimensional bar chart."""
    occurence.containing(
        *IBarChart.queryTaggedValue('directives', ())
        )

    thetaX = attr.Float(
        title=u'Theta-X',
        description=u'Fraction of dx/dz.',
        required=False)

    thetaY = attr.Float(
        title=u'Theta-Y',
        description=u'Fraction of dy/dz.',
        required=False)

    zDepth = attr.Measurement(
        title=u'Z-Depth',
        description=u'Depth of an individual series/bar.',
        required=False)

    zSpace = attr.Measurement(
        title=u'Z-Space',
        description=u'Z-Gap around a series/bar.',
        required=False)

class BarChart3D(BarChart):
    signature = IBarChart3D
    nameBase = 'BarChart3D'
    attrMapping = {'thetaX': 'theta_x', 'thetaY': 'theta_y'}


class ILinePlot(IChart):
    """A line plot."""
    occurence.containing(
        occurence.One('data', IData2D),
        occurence.ZeroOrOne('lines', ILines),
        occurence.ZeroOrOne('xValueAxis', IXValueAxis),
        occurence.ZeroOrOne('yValueAxis', IYValueAxis),
        occurence.ZeroOrOne('lineLabels', ILineLabels),
        *IChart.queryTaggedValue('directives', ())
        )

    reversePlotOrder = attr.Boolean(
        title=u'Reverse Plot Order',
        description=u'When true, the coordinate system is reversed.',
        required=False)

    lineLabelNudge = attr.Measurement(
        title=u'Line Label Nudge',
        description=u'The distance between the data point and its label.',
        required=False)

    lineLabelFormat = attr.String(
        title=u'Line Label Format',
        description=u'Formatting string for data point labels.',
        required=False)

    joinedLines = attr.Boolean(
        title=u'Joined Lines',
        description=u'When true, connect all data points with lines.',
        required=False)

class LinePlot(Chart):
    signature = ILinePlot

    factories = Chart.factories.copy()
    factories.update({
        'data': Data2D,
        'lines': Lines,
        'xValueAxis': LineXValueAxis,
        'yValueAxis': LineYValueAxis,
        'lineLabels': LineLabels,
        })

    def createChart(self, attrs):
        # Generate the chart
        chart = lineplots.LinePlot()
        for name, value in attrs.items():
            setattr(chart, name, value)
        return chart

class ILinePlot3D(ILinePlot):
    """Creates a three-dimensional line plot."""
    occurence.containing(
        *ILinePlot.queryTaggedValue('directives', ())
        )

    thetaX = attr.Float(
        title=u'Theta-X',
        description=u'Fraction of dx/dz.',
        required=False)

    thetaY = attr.Float(
        title=u'Theta-Y',
        description=u'Fraction of dy/dz.',
        required=False)

    zDepth = attr.Measurement(
        title=u'Z-Depth',
        description=u'Depth of an individual series/bar.',
        required=False)

    zSpace = attr.Measurement(
        title=u'Z-Space',
        description=u'Z-Gap around a series/bar.',
        required=False)

class LinePlot3D(LinePlot):
    signature = ILinePlot3D
    nameBase = 'LinePlot3D'
    attrMapping = {'thetaX': 'theta_x', 'thetaY': 'theta_y'}

    def createChart(self, attrs):
        # Generate the chart
        chart = lineplots.LinePlot3D()
        for name, value in attrs.items():
            setattr(chart,name, value)
        return chart



class IPieChart(IChart):
    """A pie chart."""
    occurence.containing(
        occurence.One('data', ISingleData1D),
        occurence.ZeroOrOne('slices', ISlices),
        occurence.ZeroOrOne('labels', ISimpleLabels),
        *IChart.queryTaggedValue('directives', ())
        )

    startAngle = attr.Integer(
        title=u'Start Angle',
        description=u'The start angle in the chart of the first slice '
                    u'in degrees.',
        required=False)

    direction = attr.Choice(
        title=u'Direction',
        description=u'The direction in which the pie chart will be built.',
        choices=('clockwise', 'anticlockwise'),
        required=False)

    checkLabelOverlap = attr.Boolean(
        title=u'Check Label Overlap',
        description=(u'When true, check and attempt to fix standard '
                     u'label overlaps'),
        required=False)

    pointerLabelMode = attr.Choice(
        title=u'Pointer Label Mode',
        description=(u'The location relative to the slace the label should '
                     u'be placed.'),
        choices={'none': None,
                 'leftright': 'LeftRight',
                 'leftandright': 'LeftAndRight'},
        required=False)

    sameRadii = attr.Boolean(
        title=u'Same Radii',
        description=u'When true, make x/y radii the same.',
        required=False)

    orderMode = attr.Choice(
        title=u'Order Mode',
        description=u'',
        choices=('fixed', 'alternate'),
        required=False)

    xradius = attr.Measurement(
        title=u'X-Radius',
        description=u'The radius of the X-directions',
        required=False)

    yradius = attr.Measurement(
        title=u'Y-Radius',
        description=u'The radius of the Y-directions',
        required=False)


class PieChart(Chart):
    signature = IPieChart
    chartClass = piecharts.Pie

    factories = Chart.factories.copy()
    factories.update({
        'data': SingleData1D,
        'slices': Slices,
        'labels': SimpleLabels,
        })

    def createChart(self, attrs):
        # Generate the chart
        chart = self.chartClass()
        for name, value in attrs.items():
            setattr(chart, name, value)
        return chart


class IPieChart3D(IPieChart):
    """A 3-D pie chart."""
    occurence.containing(
        occurence.One('slices', ISlices3D),
        *IChart.queryTaggedValue('directives', ())
        )

    perspective = attr.Float(
        title=u'Perspsective',
        description=u'The flattening parameter.',
        required=False)

    depth_3d = attr.Measurement(
        title=u'3-D Depth',
        description=u'The depth of the pie.',
        required=False)

    angle_3d = attr.Float(
        title=u'3-D Angle',
        description=u'The view angle in the Z-coordinate.',
        required=False)

class PieChart3D(PieChart):
    signature = IPieChart3D
    chartClass = piecharts.Pie3d

    factories = PieChart.factories.copy()
    factories.update({
        'slices': Slices3D,
        })

class ISpiderChart(IChart):
    """A spider chart."""
    occurence.containing(
        occurence.One('data', IData1D),
        occurence.ZeroOrOne('strands', IStrands),
        occurence.ZeroOrOne('strandLabels', IStrandLabels),
        occurence.ZeroOrOne('spokes', ISpokes),
        occurence.ZeroOrOne('spokeLabels', ISpokeLabels),
        occurence.ZeroOrOne('labels', ISimpleLabels),
        *IChart.queryTaggedValue('directives', ())
        )

    startAngle = attr.Integer(
        title=u'Start Angle',
        description=u'The start angle in the chart of the first strand '
                    u'in degrees.',
        required=False)

    direction = attr.Choice(
        title=u'Direction',
        description=u'The direction in which the spider chart will be built.',
        choices=('clockwise', 'anticlockwise'),
        required=False)

class SpiderChart(Chart):
    signature = ISpiderChart
    factories = Chart.factories.copy()
    factories.update({
        'data': Data1D,
        'strands': Strands,
        'strandLabels': StrandLabels,
        'spokes': Spokes,
        'spokeLabels': SpokeLabels,
        'labels': SimpleLabels,
        })

    def createChart(self, attrs):
        # Generate the chart
        chart = spider.SpiderChart()
        for name, value in attrs.items():
            setattr(chart, name, value)
        return chart



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""``pdfInclude`` Directive.
"""
__docformat__ = "reStructuredText"

import logging
import os
import subprocess
import six
from backports import tempfile

try:
    import PyPDF2
    from PyPDF2.generic import NameObject
except ImportError:
    PyPDF2 = None
from reportlab.platypus import flowables

from z3c.rml import attr, flowable, interfaces, occurence

log = logging.getLogger(__name__)

# by default False to avoid burping on
# PdfReadWarning: Multiple definitions in dictionary at byte xxx
STRICT = False

def _letter(val, base=ord('A'), radix=26):
    __traceback_info__ = val, base
    index = val - 1
    if index < 0:
        raise ValueError('Value must be greater than 0.')
    s = ''
    while True:
        val, off = divmod(index, radix)
        index = val - 1
        s = chr(base + off) + s
        if not val:
            return s

def do(cmd, cwd=None, captureOutput=True, ignoreErrors=False):
    log.debug('Command: ' + cmd)
    if captureOutput:
        stdout = stderr = subprocess.PIPE
    else:
        stdout = stderr = None
    p = subprocess.Popen(
        cmd, stdout=stdout, stderr=stderr,
        shell=True, cwd=cwd)
    stdout, stderr = p.communicate()
    if stdout is None:
        stdout = "See output above"
    if stderr is None:
        stderr = "See output above"
    if p.returncode != 0 and not ignoreErrors:
        log.error(u'An error occurred while running command: %s' % cmd)
        log.error('Error Output: \n%s' % stderr)
        raise ValueError('Shell Process had non-zero error code: {}. \n'
                         'Stdout: {}\n'
                         'StdErr: {}'.format(p.returncode, stdout, stderr))
    log.debug('Output: \n%s' % stdout)
    return stdout


class ConcatenationPostProcessor(object):

    def __init__(self):
        self.operations = []

    def process(self, inputFile1):
        input1 = PyPDF2.PdfFileReader(inputFile1, strict=STRICT)
        merger = PyPDF2.PdfFileMerger(strict=STRICT)
        merger.output._info.getObject().update(input1.documentInfo)

        merger.append(inputFile1)

        for start_page, inputFile2, page_ranges, num_pages in self.operations:
            # Remove blank pages, that we reserved in IncludePdfPagesFlowable
            # and insert real pdf here
            del merger.pages[start_page:start_page + num_pages]
            curr_page = start_page
            for page_range in page_ranges:
                prs, pre = page_range
                merger.merge(
                    curr_page, inputFile2, pages=(prs, pre),
                    import_bookmarks=False)

        outputFile = six.BytesIO()
        merger.write(outputFile)
        return outputFile


class PdfTkConcatenationPostProcessor(object):

    EXECUTABLE = 'pdftk'
    PRESERVE_OUTLINE = True

    def __init__(self):
        self.operations = []

    def _process(self, inputFile1, dir):
        file_path = os.path.join(dir, 'A.pdf')
        with open(file_path, 'wb') as file:
            file.write(inputFile1.read())

        file_map = {'A': file_path}
        file_id = 2
        merges = []

        curr_page = 0
        for start_page, inputFile2, page_ranges, num_pages in self.operations:
            # Catch up with the main file.
            if curr_page < start_page:
                # Convert curr_page to human counting, start_page is okay,
                # since pdftk is upper-bound inclusive.
                merges.append('A%i-%i' % (curr_page+1, start_page))
            curr_page = start_page + num_pages

            # Store file.
            file_letter = _letter(file_id)
            file_path = os.path.join(dir, file_letter+'.pdf')
            inputFile2.seek(0)
            with open(file_path, 'wb') as file:
                file.write(inputFile2.read())
            file_map[file_letter] = file_path
            file_id += 1

            for (prs, pre) in page_ranges:
                # pdftk uses lower and upper bound inclusive.
                merges.append('%s%i-%i' % (file_letter, prs+1, pre))

        mergedFile = os.path.join(dir, 'merged.pdf')
        do('%s %s cat %s output %s' % (
            self.EXECUTABLE,
            ' '.join('%s="%s"' % (l, p) for l, p in file_map.items()),
            ' '.join(merges),
            mergedFile))

        if not self.PRESERVE_OUTLINE:
            with open(mergedFile, 'rb') as file:
                return six.BytesIO(file.read())

        outputFile = os.path.join(dir, 'output.pdf')
        do('%s %s/A.pdf dump_data > %s/in.info' % (
            self.EXECUTABLE, dir, dir))
        do('%s %s update_info %s/in.info output %s' % (
            self.EXECUTABLE, mergedFile, dir, outputFile))

        with open(outputFile, 'rb') as file:
            return six.BytesIO(file.read())

    def process(self, inputFile1):
        with tempfile.TemporaryDirectory() as tmpdirname:
            return self._process(inputFile1, tmpdirname)


class IncludePdfPagesFlowable(flowables.Flowable):

    def __init__(self, pdf_file, pages, concatprocessor,
                 included_on_first_page):
        flowables.Flowable.__init__(self)
        self.pdf_file = pdf_file
        self.proc = concatprocessor
        self.pages = pages
        self.included_on_first_page = included_on_first_page

        if self.included_on_first_page:
            self.width = 0
            self.height = 0
        else:
            self.width = 10<<32
            self.height = 10<<32

    def draw(self):
        if self.included_on_first_page:
            self.split(None, None)

    def split(self, availWidth, availheight):
        pages = self.pages
        if not pages:
            pdf = PyPDF2.PdfFileReader(self.pdf_file, strict=STRICT)
            pages = [(0, pdf.getNumPages())]

        num_pages = sum(pr[1]-pr[0] for pr in pages)

        start_page = self.canv.getPageNumber()
        if self.included_on_first_page:
            start_page -= 1
        self.proc.operations.append(
            (start_page, self.pdf_file, pages, num_pages))

        # Insert blank pages instead of pdf for now, to correctly number the
        # pages. We will replace these blank pages with included PDF in
        # ConcatenationPostProcessor.
        result = []
        for i in range(num_pages):
            # Add empty spacer so platypus don't complain about too many empty
            # pages
            result.append(flowables.Spacer(0, 0))
            result.append(flowables.PageBreak())
        return result


class IIncludePdfPages(interfaces.IRMLDirectiveSignature):
    """Inserts a set of pages from a given PDF."""

    filename = attr.File(
        title=u'Path to file',
        description=u'The pdf file to include.',
        required=True)

    pages = attr.IntegerSequence(
        title=u'Pages',
        description=u'A list of pages to insert.',
        numberingStartsAt=1,
        required=False)


class IncludePdfPages(flowable.Flowable):
    signature = IIncludePdfPages

    ConcatenationPostProcessorFactory = ConcatenationPostProcessor

    def getProcessor(self):
        manager = attr.getManager(self, interfaces.IPostProcessorManager)
        procs = dict(manager.postProcessors)
        if 'CONCAT' not in procs:
            log.debug(
                'Using concetation post-processor: %s',
                self.ConcatenationPostProcessorFactory)
            proc = self.ConcatenationPostProcessorFactory()
            manager.postProcessors.append(('CONCAT', proc))
            return proc
        return procs['CONCAT']

    def process(self):
        if PyPDF2 is None:
            raise Exception(
                'PyPDF2 is not installed, so this feature is not available.')
        args = dict(self.getAttributeValues())
        proc = self.getProcessor()
        self.parent.flow.append(
            IncludePdfPagesFlowable(
                args['filename'], args.get('pages'), proc, not self.parent.flow
            ))


flowable.Flow.factories['includePdfPages'] = IncludePdfPages
flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('includePdfPages', IIncludePdfPages),)
    )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML to PDF Converter Interfaces
"""
__docformat__ = "reStructuredText"
import collections
import logging
import reportlab.lib.enums
import zope.interface
import zope.schema

from z3c.rml.occurence import ZeroOrMore, ZeroOrOne, OneOrMore

JOIN_CHOICES = collections.OrderedDict([
    ('round', 1), ('mitered', 0), ('bevelled', 2)
])

CAP_CHOICES = collections.OrderedDict([
    ('default', 0), ('butt', 0), ('round', 1), ('square', 2)
])

ALIGN_CHOICES = {
    'left': reportlab.lib.enums.TA_LEFT,
    'right': reportlab.lib.enums.TA_RIGHT,
    'center': reportlab.lib.enums.TA_CENTER,
    'centre': reportlab.lib.enums.TA_CENTER,
    'justify': reportlab.lib.enums.TA_JUSTIFY}

ALIGN_TEXT_CHOICES = collections.OrderedDict([
    ('left', 'LEFT'), ('right', 'RIGHT'),
    ('center', 'CENTER'), ('centre', 'CENTER'),
    ('decimal', 'DECIMAL')
])

VALIGN_TEXT_CHOICES = {
    'top': 'TOP', 'middle': 'MIDDLE', 'bottom': 'BOTTOM'}

SPLIT_CHOICES = ('splitfirst', 'splitlast')

WORD_WRAP_CHOICES = ('CJK', 'RTL', 'LTR')

TEXT_TRANSFORM_CHOICES = ('none', 'uppercase', 'lowercase', 'capitalize')

TEXT_DECORATION_CHOICES = ('underline', 'strike')

UNDERLINE_KIND_CHOICES = ('single', 'double', 'triple')

STRIKE_KIND_CHOICES = ('single', 'double', 'triple')

BULLET_ANCHOR_CHOICES = ('start', 'middle', 'end', 'numeric')

LIST_FORMATS = ('I', 'i', '123',  'ABC', 'abc')

ORDERED_LIST_TYPES = ('I', 'i', '1', 'A', 'a', 'l', 'L', 'O', 'o', 'R', 'r')

UNORDERED_BULLET_VALUES = (
    'bulletchar', 'bullet', 'circle', 'square', 'disc', 'diamond',
    'rarrowhead')

LOG_LEVELS = collections.OrderedDict([
    ('DEBUG', logging.DEBUG),
    ('INFO', logging.INFO),
    ('WARNING', logging.WARNING),
    ('ERROR', logging.ERROR),
    ('CRITICAL', logging.CRITICAL),
])


class IRML2PDF(zope.interface.Interface):
    """This is the main public API of z3c.rml"""

    def parseString(xml):
        """Parse an XML string and convert it to PDF.

        The output is a ``StringIO`` object.
        """

    def go(xmlInputName, outputFileName=None, outDir=None, dtdDir=None):
        """Convert RML 2 PDF.

        The generated file will be located in the ``outDir`` under the name
        ``outputFileName``.
        """

class IManager(zope.interface.Interface):
    """A manager of all document-global variables."""
    names = zope.interface.Attribute("Names dict")
    styles = zope.interface.Attribute("Styles dict")
    colors = zope.interface.Attribute("Colors dict")

class IPostProcessorManager(zope.interface.Interface):
    """Manages all post processors"""

    postProcessors = zope.interface.Attribute(
        "List of tuples of the form: (name, processor)")

class ICanvasManager(zope.interface.Interface):
    """A manager for the canvas."""
    canvas = zope.interface.Attribute("Canvas")

class IRMLDirectiveSignature(zope.interface.Interface):
    """The attribute and sub-directives signature of the current
    RML directive."""


class IRMLDirective(zope.interface.Interface):
    """A directive in RML extracted from an Element Tree element."""

    signature = zope.schema.Field(
        title=u'Signature',
        description=(u'The signature of the RML directive.'),
        required=True)

    parent = zope.schema.Field(
        title=u'Parent RML Element',
        description=u'The parent in the RML element hierarchy',
        required=True,)

    element = zope.schema.Field(
        title=u'Element',
        description=(u'The Element Tree element from which the data '
                     u'is retrieved.'),
        required=True)

    def getAttributeValues(ignore=None, select=None, includeMissing=False):
        """Return a list of name-value-tuples based on the signature.

        If ``ignore`` is specified, all attributes are returned except the
        ones listed in the argument. The values of the sequence are the
        attribute names.

        If ``select`` is specified, only attributes listed in the argument are
        returned. The values of the sequence are the attribute names.

        If ``includeMissing`` is set to true, then even missing attributes are
        included in the value list.
        """

    def processSubDirectives(self):
        """Process all sub-directives."""

    def process(self):
        """Process the directive.

        The main task for this method is to interpret the available data and
        to make the corresponding calls in the Reportlab document.

        This call should also process all sub-directives and process them.
        """


class IDeprecated(zope.interface.Interface):
    """Mark an attribute as being compatible."""

    deprecatedName = zope.schema.TextLine(
        title=u'Name',
        description=u'The name of the original attribute.',
        required=True)

    deprecatedReason = zope.schema.Text(
        title=u'Reason',
        description=u'The reason the attribute has been deprecated.',
        required=False)


class IDeprecatedDirective(zope.interface.interfaces.IInterface):
    """A directive that is deprecated."""



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""``ul``, ``ol``, and ``li`` directives.
"""
__docformat__ = "reStructuredText"
import copy
import reportlab.lib.styles
import reportlab.platypus
import zope.schema
from reportlab.platypus import flowables

from z3c.rml import attr, directive, flowable, interfaces, occurence, stylesheet


class IListItem(stylesheet.IMinimalListStyle, flowable.IFlow):
    """A list item in an ordered or unordered list."""

    style = attr.Style(
        title=u'Style',
        description=u'The list style that is applied to the list.',
        required=False)

class ListItem(flowable.Flow):
    signature = IListItem
    klass = reportlab.platypus.ListItem
    attrMapping = {}

    styleAttributes = zope.schema.getFieldNames(stylesheet.IMinimalListStyle)

    def processStyle(self, style):
        attrs = self.getAttributeValues(select=self.styleAttributes)
        if attrs or not hasattr(style, 'value'):
            style = copy.deepcopy(style)
            # Sigh, this is needed since unordered list items expect the value.
            style.value = style.start
            for name, value in attrs:
                setattr(style, name, value)
        return style

    def process(self):
        self.processSubDirectives()
        args = dict(self.getAttributeValues(ignore=self.styleAttributes))
        if 'style' not in args:
            args['style'] = self.parent.baseStyle
        args['style'] = self.processStyle(args['style'])
        li = self.klass(self.flow, **args)
        self.parent.flow.append(li)


class IOrderedListItem(IListItem):
    """An ordered list item."""

    value = attr.Integer(
        title=u'Bullet Value',
        description=u'The counter value.',
        required=False)

class OrderedListItem(ListItem):
    signature = IOrderedListItem


class IUnorderedListItem(IListItem):
    """An ordered list item."""

    value = attr.Choice(
        title=u'Bullet Value',
        description=u'The type of bullet character.',
        choices=interfaces.UNORDERED_BULLET_VALUES,
        required=False)

class UnorderedListItem(ListItem):
    signature = IUnorderedListItem

    styleAttributes = ListItem.styleAttributes + ['value']


class IListBase(stylesheet.IBaseListStyle):

    style = attr.Style(
        title=u'Style',
        description=u'The list style that is applied to the list.',
        required=False)

class ListBase(directive.RMLDirective):
    klass = reportlab.platypus.ListFlowable
    factories = {'li': ListItem}
    attrMapping = {}

    styleAttributes = zope.schema.getFieldNames(stylesheet.IBaseListStyle)

    def __init__(self, *args, **kw):
        super(ListBase, self).__init__(*args, **kw)
        self.flow = []

    def processStyle(self, style):
        attrs = self.getAttributeValues(
            select=self.styleAttributes, attrMapping=self.attrMapping)
        if attrs:
            style = copy.deepcopy(style)
            for name, value in attrs:
                setattr(style, name, value)
        return style

    def process(self):
        args = dict(self.getAttributeValues(
                ignore=self.styleAttributes, attrMapping=self.attrMapping))
        if 'style' not in args:
            args['style'] = reportlab.lib.styles.ListStyle('List')
        args['style'] = self.baseStyle = self.processStyle(args['style'])
        self.processSubDirectives()
        li = self.klass(self.flow, **args)
        self.parent.flow.append(li)


class IOrderedList(IListBase):
    """An ordered list."""
    occurence.containing(
        occurence.ZeroOrMore('li', IOrderedListItem),
        )

    bulletType = attr.Choice(
        title=u'Bullet Type',
        description=u'The type of bullet formatting.',
        choices=interfaces.ORDERED_LIST_TYPES,
        doLower=False,
        required=False)

class OrderedList(ListBase):
    signature = IOrderedList
    factories = {'li': OrderedListItem}

    styleAttributes = ListBase.styleAttributes + ['bulletType']


class IUnorderedList(IListBase):
    """And unordered list."""
    occurence.containing(
        occurence.ZeroOrMore('li', IUnorderedListItem),
        )

    value = attr.Choice(
        title=u'Bullet Value',
        description=u'The type of bullet character.',
        choices=interfaces.UNORDERED_BULLET_VALUES,
        default='disc',
        required=False)

class UnorderedList(ListBase):
    signature = IUnorderedList
    attrMapping = {'value': 'start'}
    factories = {'li': UnorderedListItem}

    def getAttributeValues(self, *args, **kw):
        res = super(UnorderedList, self).getAttributeValues(*args, **kw)
        res.append(('bulletType', 'bullet'))
        return res

flowable.Flow.factories['ol'] = OrderedList
flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('ol', IOrderedList),)
    )

flowable.Flow.factories['ul'] = UnorderedList
flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('ul', IUnorderedList),)
    )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Style Related Element Processing
"""
import reportlab.platypus.flowables


class TestFlowable(reportlab.platypus.flowables.Flowable):

    def __init__(self, text):
        self.text = text

    def wrap(self, *args):
        return (400, 100)

    def draw(self):
        self.canv.saveState()
        self.canv.rect(0, 0, 400, 100)
        self.canv.drawString(0, 50, self.text)
        self.canv.restoreState()
